@@ -27,7 +27,6 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netinet/ether.h>
-#include "wireless-helper.h"
 #include <ctype.h>
 
 #include <glib/gi18n.h>
@@ -45,7 +44,6 @@
 #include "applet-device-wifi.h"
 #include "ap-menu-item.h"
 #include "utils.h"
-#include "gnome-keyring-md5.h"
 #include "wireless-dialog.h"
 
 #define ACTIVE_AP_TAG "active-ap"
@@ -422,22 +420,21 @@ wireless_menu_item_activate (GtkMenuItem *item, gpointer user_data)
 	                                  user_data);
 }
 
-#define AP_HASH_LEN 16
-
 struct dup_data {
-	NMDevice * device;
-	GtkWidget * found;
-	guchar * hash;
+	NMDevice *device;
+	NMNetworkMenuItem *found;
+	char *hash;
 };
 
 static void
-find_duplicate (GtkWidget * widget, gpointer user_data)
+find_duplicate (gpointer d, gpointer user_data)
 {
 	struct dup_data * data = (struct dup_data *) user_data;
 	NMDevice *device;
-	const guchar * hash;
-	guint32 hash_len = 0;
+	const char *hash;
+	GtkWidget *widget = GTK_WIDGET (d);
 
+	g_assert (d && widget);
 	g_return_if_fail (data);
 	g_return_if_fail (data->hash);
 
@@ -448,12 +445,9 @@ find_duplicate (GtkWidget * widget, gpointer user_data)
 	if (NM_DEVICE (device) != data->device)
 		return;
 
-	hash = nm_network_menu_item_get_hash (NM_NETWORK_MENU_ITEM (widget), &hash_len);
-	if (hash == NULL || hash_len != AP_HASH_LEN)
-		return;
-
-	if (memcmp (hash, data->hash, AP_HASH_LEN) == 0)
-		data->found = widget;
+	hash = nm_network_menu_item_get_hash (NM_NETWORK_MENU_ITEM (widget));
+	if (hash && (strcmp (hash, data->hash) == 0))
+		data->found = NM_NETWORK_MENU_ITEM (widget);
 }
 
 static GSList *
@@ -472,44 +466,36 @@ filter_connections_for_access_point (GSList *connections, NMDeviceWifi *device,
 }
 
 static NMNetworkMenuItem *
-add_new_ap_item (NMDeviceWifi *device,
-                 NMAccessPoint *ap,
-                 struct dup_data *dup_data,
-                 NMAccessPoint *active_ap,
-                 NMConnection *active,
-                 GSList *connections,
-                 GtkWidget *menu,
-                 NMApplet *applet)
+create_new_ap_item (NMDeviceWifi *device,
+                    NMAccessPoint *ap,
+                    struct dup_data *dup_data,
+                    GSList *connections,
+                    NMApplet *applet)
 {
 	WirelessMenuItemInfo *info;
-	GtkWidget *foo;
 	GSList *iter;
 	NMNetworkMenuItem *item = NULL;
 	GSList *ap_connections = NULL;
 	const GByteArray *ssid;
-	guint8 strength;
 	guint32 dev_caps;
 
 	ap_connections = filter_connections_for_access_point (connections, device, ap);
 
-	foo = nm_network_menu_item_new (applet->encryption_size_group,
-	                                dup_data->hash, AP_HASH_LEN);
-	item = NM_NETWORK_MENU_ITEM (foo);
+	item = NM_NETWORK_MENU_ITEM (nm_network_menu_item_new (dup_data->hash,
+	                                                       !!g_slist_length (ap_connections)));
+	gtk_image_menu_item_set_always_show_image (GTK_IMAGE_MENU_ITEM (item), TRUE);
 
 	ssid = nm_access_point_get_ssid (ap);
 	nm_network_menu_item_set_ssid (item, (GByteArray *) ssid);
 
-	strength = nm_access_point_get_strength (ap);
-	nm_network_menu_item_set_strength (item, strength);
-
 	dev_caps = nm_device_wifi_get_capabilities (device);
+	nma_icon_check_and_load ("nm-adhoc", &applet->adhoc_icon, applet);
 	nm_network_menu_item_set_detail (item, ap, applet->adhoc_icon, dev_caps);
+	nm_network_menu_item_best_strength (item, nm_access_point_get_strength (ap), applet);
 	nm_network_menu_item_add_dupe (item, ap);
 
 	g_object_set_data (G_OBJECT (item), "device", NM_DEVICE (device));
 
-	gtk_menu_shell_append (GTK_MENU_SHELL (menu), GTK_WIDGET (item));
-
 	/* If there's only one connection, don't show the submenu */
 	if (g_slist_length (ap_connections) > 1) {
 		GtkWidget *submenu;
@@ -560,84 +546,64 @@ add_new_ap_item (NMDeviceWifi *device,
 		                       0);
 	}
 
-	gtk_widget_show_all (GTK_WIDGET (item));
-
 	g_slist_free (ap_connections);
 	return item;
 }
 
-static void
-add_one_ap_menu_item (NMDeviceWifi *device,
+static NMNetworkMenuItem *
+get_menu_item_for_ap (NMDeviceWifi *device,
                       NMAccessPoint *ap,
                       GSList *connections,
-                      NMAccessPoint *active_ap,
-                      NMConnection *active,
-                      GtkWidget *menu,
+                      GSList *menu_list,
                       NMApplet *applet)
 {
 	const GByteArray *ssid;
 	struct dup_data dup_data = { NULL, NULL };
-	NMNetworkMenuItem *item = NULL;
 
 	/* Don't add BSSs that hide their SSID */
 	ssid = nm_access_point_get_ssid (ap);
 	if (!ssid || nm_utils_is_empty_ssid (ssid->data, ssid->len))
-		return;
+		return NULL;
 
+	/* Find out if this AP is a member of a larger network that all uses the
+	 * same SSID and security settings.  If so, we'll already have a menu item
+	 * for this SSID, so just update that item's strength and add this AP to
+	 * menu item's duplicate list.
+	 */
 	dup_data.found = NULL;
 	dup_data.hash = g_object_get_data (G_OBJECT (ap), "hash");
-	if (!dup_data.hash)
-		return;
+	g_return_val_if_fail (dup_data.hash != NULL, NULL);
+
 	dup_data.device = NM_DEVICE (device);
-	gtk_container_foreach (GTK_CONTAINER (menu),
-	                       find_duplicate,
-	                       &dup_data);
+	g_slist_foreach (menu_list, find_duplicate, &dup_data);
 
 	if (dup_data.found) {
-		gint8 strength = nm_access_point_get_strength (ap);
-
-		item = NM_NETWORK_MENU_ITEM (dup_data.found);
-
-		/* Just update strength if greater than what's there */
-		if (nm_network_menu_item_get_strength (item) < strength)
-			nm_network_menu_item_set_strength (item, strength);
-
-		nm_network_menu_item_add_dupe (item, ap);
-	} else {
-		item = add_new_ap_item (device, ap, &dup_data, active_ap, active, connections, menu, applet);
+		nm_network_menu_item_best_strength (dup_data.found, nm_access_point_get_strength (ap), applet);
+		nm_network_menu_item_add_dupe (dup_data.found, ap);
+		return NULL;
 	}
 
-	if (!active_ap)
-		return;
-
-	g_signal_handlers_block_matched (item, G_SIGNAL_MATCH_FUNC, 0, 0, NULL,
-	                                 G_CALLBACK (wireless_menu_item_activate), NULL);
-
-	if (nm_network_menu_item_find_dupe (item, active_ap))
-		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (item), TRUE);
-
-	g_signal_handlers_unblock_matched (item, G_SIGNAL_MATCH_FUNC, 0, 0, NULL,
-	                                   G_CALLBACK (wireless_menu_item_activate), NULL);
+	return create_new_ap_item (device, ap, &dup_data, connections, applet);
 }
 
 static gint
-sort_wireless_networks (gconstpointer tmpa,
-                        gconstpointer tmpb)
+sort_by_name (gconstpointer tmpa, gconstpointer tmpb)
 {
-	NMAccessPoint * a = NM_ACCESS_POINT (tmpa);
-	NMAccessPoint * b = NM_ACCESS_POINT (tmpb);
-	const GByteArray * a_ssid;
-	const GByteArray * b_ssid;
-	NM80211Mode a_mode, b_mode;
+	NMNetworkMenuItem *a = NM_NETWORK_MENU_ITEM (tmpa);
+	NMNetworkMenuItem *b = NM_NETWORK_MENU_ITEM (tmpb);
+	const char *a_ssid, *b_ssid;
+	gboolean a_adhoc, b_adhoc;
 	int i;
 
 	if (a && !b)
 		return 1;
-	if (b && !a)
+	else if (!a && b)
 		return -1;
+	else if (a == b)
+		return 0;
 
-	a_ssid = nm_access_point_get_ssid (a);
-	b_ssid = nm_access_point_get_ssid (b);
+	a_ssid = nm_network_menu_item_get_ssid (a);
+	b_ssid = nm_network_menu_item_get_ssid (b);
 
 	if (a_ssid && !b_ssid)
 		return 1;
@@ -645,33 +611,67 @@ sort_wireless_networks (gconstpointer tmpa,
 		return -1;
 
 	if (a_ssid && b_ssid) {
-		/* Can't use string compares because SSIDs are byte arrays and
-		 * may legally contain embedded NULLs.
-		 */
-		for (i = 0; i < MIN(a_ssid->len, b_ssid->len); i++) {
-			if (tolower(a_ssid->data[i]) > tolower(b_ssid->data[i]))
-				return 1;
-			else if (tolower(b_ssid->data[i]) > tolower(a_ssid->data[i]))
-				return -1;
-		}
-
-		if (a_ssid->len > b_ssid->len)
-			return 1;
-		if (b_ssid->len > a_ssid->len)
-			return -1;
+		i = g_ascii_strcasecmp (a_ssid, b_ssid);
+		if (i != 0)
+			return i;
 	}
 
-	a_mode = nm_access_point_get_mode (a);
-	b_mode = nm_access_point_get_mode (b);
-	if (a_mode != b_mode) {
-		if (a_mode == NM_802_11_MODE_INFRA)
-			return 1;
+	/* If the names are the same, sort infrastructure APs first */
+	a_adhoc = nm_network_menu_item_get_is_adhoc (a);
+	b_adhoc = nm_network_menu_item_get_is_adhoc (b);
+	if (a_adhoc && !b_adhoc)
+		return 1;
+	else if (!a_adhoc && b_adhoc)
 		return -1;
-	}
 
 	return 0;
 }
 
+/* Sort menu items for the top-level menu:
+ * 1) whether there's a saved connection or not
+ *    a) sort alphabetically within #1
+ * 2) encrypted without a saved connection
+ * 3) unencrypted without a saved connection
+ */
+static gint
+sort_toplevel (gconstpointer tmpa, gconstpointer tmpb)
+{
+	NMNetworkMenuItem *a = NM_NETWORK_MENU_ITEM (tmpa);
+	NMNetworkMenuItem *b = NM_NETWORK_MENU_ITEM (tmpb);
+	gboolean a_fave, b_fave;
+
+	if (a && !b)
+		return 1;
+	else if (!a && b)
+		return -1;
+	else if (a == b)
+		return 0;
+
+	a_fave = nm_network_menu_item_get_has_connections (a);
+	b_fave = nm_network_menu_item_get_has_connections (b);
+
+	/* Items with a saved connection first */
+	if (a_fave && !b_fave)
+		return -1;
+	else if (!a_fave && b_fave)
+		return 1;
+	else if (!a_fave && !b_fave) {
+		gboolean a_enc = nm_network_menu_item_get_is_encrypted (a);
+		gboolean b_enc = nm_network_menu_item_get_is_encrypted (b);
+
+		/* If neither item has a saved connection, sort by encryption */
+		if (a_enc && !b_enc)
+			return -1;
+		else if (!a_enc && b_enc)
+			return 1;
+	}
+
+	/* For all other cases (both have saved connections, both are encrypted, or
+	 * both are unencrypted) just sort by name.
+	 */
+	return sort_by_name (a, b);
+}
+
 static void
 wireless_add_menu_item (NMDevice *device,
                         guint32 n_devices,
@@ -681,22 +681,18 @@ wireless_add_menu_item (NMDevice *device,
 {
 	NMDeviceWifi *wdev;
 	char *text;
-	GtkWidget *item;
 	const GPtrArray *aps;
 	int i;
 	NMAccessPoint *active_ap = NULL;
-	GSList *connections = NULL, *all, *sorted_aps = NULL, *iter;
-	GtkWidget *label;
-	char *bold_text;
+	GSList *connections = NULL, *all, *iter;
 	gboolean wireless_enabled = TRUE;
+	GSList *menu_items = NULL;  /* All menu items we'll be adding */
+	NMNetworkMenuItem *item, *active_item = NULL;
+	GtkWidget *widget;
 
 	wdev = NM_DEVICE_WIFI (device);
 	aps = nm_device_wifi_get_access_points (wdev);
 
-	all = applet_get_all_connections (applet);
-	connections = utils_filter_connections_for_device (device, all);
-	g_slist_free (all);
-
 	if (n_devices > 1) {
 		char *desc;
 
@@ -712,62 +708,139 @@ wireless_add_menu_item (NMDevice *device,
 	} else
 		text = g_strdup (ngettext ("Wireless Network", "Wireless Networks", aps ? aps->len : 0));
 
-	item = gtk_menu_item_new_with_mnemonic (text);
+	widget = applet_menu_item_create_device_item_helper (device, applet, text);
 	g_free (text);
 
-	label = gtk_bin_get_child (GTK_BIN (item));
-	bold_text = g_markup_printf_escaped ("<span weight=\"bold\">%s</span>",
-	                                     gtk_label_get_text (GTK_LABEL (label)));
-	gtk_label_set_markup (GTK_LABEL (label), bold_text);
-	g_free (bold_text);
+	gtk_widget_set_sensitive (widget, FALSE);
+	gtk_menu_shell_append (GTK_MENU_SHELL (menu), widget);
+	gtk_widget_show (widget);
 
-	gtk_widget_set_sensitive (item, FALSE);
-	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
-	gtk_widget_show (item);
+	all = applet_get_all_connections (applet);
+	connections = utils_filter_connections_for_device (device, all);
+	g_slist_free (all);
+
+	/* Add the active AP if we're connected to something and the device is available */
+	if (!nma_menu_device_check_unusable (device)) {
+		active_ap = nm_device_wifi_get_active_access_point (wdev);
+		if (active_ap) {
+			active_item = item = get_menu_item_for_ap (wdev, active_ap, connections, NULL, applet);
+			if (item) {
+				nm_network_menu_item_set_active (item, TRUE);
+				menu_items = g_slist_append (menu_items, item);
+
+				gtk_menu_shell_append (GTK_MENU_SHELL (menu), GTK_WIDGET (item));
+				gtk_widget_show_all (GTK_WIDGET (item));
+			}
+		}
+	}
 
 	/* Notify user of unmanaged or unavailable device */
 	wireless_enabled = nm_client_wireless_get_enabled (applet->nm_client);
-	item = nma_menu_device_check_unusable (device, wireless_enabled ? NULL : _("wireless is disabled"));
-	if (item) {
-		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
-		gtk_widget_show (item);
-		goto out;
+	widget = nma_menu_device_get_menu_item (device, applet, wireless_enabled ? NULL : _("wireless is disabled"));
+	if (widget) {
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), widget);
+		gtk_widget_show (widget);
 	}
 
-	active_ap = nm_device_wifi_get_active_access_point (wdev);
+	/* If disabled or rfkilled or whatever, nothing left to do */
+	if (nma_menu_device_check_unusable (device))
+		goto out;
 
-	/* Add all networks in our network list to the menu */
-	for (i = 0; aps && (i < aps->len); i++)
-		sorted_aps = g_slist_append (sorted_aps, g_ptr_array_index (aps, i));
+	/* Create menu items for the rest of the APs */
+	for (i = 0; aps && (i < aps->len); i++) {
+		NMAccessPoint *ap = g_ptr_array_index (aps, i);
 
-	sorted_aps = g_slist_sort (sorted_aps, sort_wireless_networks);
-	for (iter = sorted_aps; iter; iter = g_slist_next (iter))
-		add_one_ap_menu_item (wdev, NM_ACCESS_POINT (iter->data), connections, active_ap, active, menu, applet);
+		item = get_menu_item_for_ap (wdev, ap, connections, menu_items, applet);
+		if (item)
+			menu_items = g_slist_append (menu_items, item);
+	}
+
+	/* Now remove the active AP item from the list, as we've already dealt with
+	 * it.  (Needed it when creating menu items for the rest of the APs though
+	 * to ensure duplicate APs are handled correctly)
+	 */
+	if (active_item)
+		menu_items = g_slist_remove (menu_items, active_item);
+
+	/* Sort all the rest of the menu items for the top-level menu */
+	menu_items = g_slist_sort (menu_items, sort_toplevel);
+
+	if (g_slist_length (menu_items)) {
+		GSList *submenu_items = NULL;
+		GSList *topmenu_items = NULL;
+		guint32 num_for_toplevel = 5;
+
+		applet_menu_item_add_complex_separator_helper (menu, applet, _("Available"), -1);
+
+		if (g_slist_length (menu_items) == (num_for_toplevel + 1))
+			num_for_toplevel++;
+
+		/* Add the first 5 APs (or 6 if there are only 6 total) from the sorted
+		 * toplevel list.
+		 */
+		for (iter = menu_items; iter && num_for_toplevel; iter = g_slist_next (iter)) {
+			topmenu_items = g_slist_append (topmenu_items, iter->data);
+			num_for_toplevel--;
+			submenu_items = iter->next;
+		}
+		topmenu_items = g_slist_sort (topmenu_items, sort_by_name);
+
+		for (iter = topmenu_items; iter; iter = g_slist_next (iter)) {
+			gtk_menu_shell_append (GTK_MENU_SHELL (menu), GTK_WIDGET (iter->data));
+			gtk_widget_show_all (GTK_WIDGET (iter->data));
+		}
+		g_slist_free (topmenu_items);
+		topmenu_items = NULL;
+
+		/* If there are any submenu items, make a submenu for those */
+		if (submenu_items) {
+			GtkWidget *subitem, *submenu;
+			GSList *sorted_subitems;
+
+			subitem = gtk_menu_item_new_with_mnemonic (_("More networks"));
+			submenu = gtk_menu_new ();
+			gtk_menu_item_set_submenu (GTK_MENU_ITEM (subitem), submenu);
+
+			/* Sort the subitems alphabetically */
+			sorted_subitems = g_slist_copy (submenu_items);
+			sorted_subitems = g_slist_sort (sorted_subitems, sort_by_name);
+
+			/* And add the rest to the submenu */
+			for (iter = sorted_subitems; iter; iter = g_slist_next (iter))
+				gtk_menu_shell_append (GTK_MENU_SHELL (submenu), GTK_WIDGET (iter->data));
+			g_slist_free (sorted_subitems);
+
+			gtk_menu_shell_append (GTK_MENU_SHELL (menu), subitem);
+			gtk_widget_show_all (subitem);
+		}
+	}
 
 out:
+	g_slist_free (menu_items);
 	g_slist_free (connections);
-	g_slist_free (sorted_aps);
 }
 
-static gboolean
-add_seen_bssid (NMAGConfConnection *gconf_connection, NMAccessPoint *ap)
+static void
+add_seen_bssid (NMSettingsConnectionInterface *connection, NMAccessPoint *ap)
 {
-	NMConnection *connection;
 	NMSettingWireless *s_wireless;
 	const char *bssid;
 
-	connection = nm_exported_connection_get_connection (NM_EXPORTED_CONNECTION (gconf_connection));
-	g_return_val_if_fail (NM_IS_CONNECTION (connection), FALSE);
+	if (!NMA_GCONF_CONNECTION (connection))
+		return;
 
-	s_wireless = NM_SETTING_WIRELESS (nm_connection_get_setting (connection, NM_TYPE_SETTING_WIRELESS));
+	s_wireless = NM_SETTING_WIRELESS (nm_connection_get_setting (NM_CONNECTION (connection), NM_TYPE_SETTING_WIRELESS));
 	if (!s_wireless)
-		return FALSE;
+		return;
 
 	bssid = nm_access_point_get_hw_address (ap);
 	if (!bssid || !utils_ether_addr_valid (ether_aton (bssid)))
-		return FALSE;
+		return;
 
-	return nm_setting_wireless_add_seen_bssid (s_wireless, bssid);
+	if (nm_setting_wireless_add_seen_bssid (s_wireless, bssid)) {
+		/* Ignore secrets since we don't have any here and we're just adding a BSSID */
+		nma_gconf_connection_update (NMA_GCONF_CONNECTION (connection), TRUE);
+	}
 }
 
 static void
@@ -775,8 +848,7 @@ notify_active_ap_changed_cb (NMDeviceWifi *device,
                              GParamSpec *pspec,
                              NMApplet *applet)
 {
-	NMAGConfConnection *gconf_connection;
-	NMConnection *connection;
+	NMSettingsConnectionInterface *connection;
 	NMSettingWireless *s_wireless;
 	NMAccessPoint *new;
 	const GByteArray *ssid;
@@ -788,14 +860,11 @@ notify_active_ap_changed_cb (NMDeviceWifi *device,
 	if (!new || (state != NM_DEVICE_STATE_ACTIVATED))
 		return;
 
-	gconf_connection = applet_get_exported_connection_for_device (NM_DEVICE (device), applet);
-	if (!gconf_connection)
+	connection = applet_get_exported_connection_for_device (NM_DEVICE (device), applet);
+	if (!connection)
 		return;
 
-	connection = nm_exported_connection_get_connection (NM_EXPORTED_CONNECTION (gconf_connection));
-	g_return_if_fail (NM_IS_CONNECTION (connection));
-
-	s_wireless = NM_SETTING_WIRELESS (nm_connection_get_setting (connection, NM_TYPE_SETTING_WIRELESS));
+	s_wireless = NM_SETTING_WIRELESS (nm_connection_get_setting (NM_CONNECTION (connection), NM_TYPE_SETTING_WIRELESS));
 	if (!s_wireless)
 		return;
 
@@ -803,80 +872,22 @@ notify_active_ap_changed_cb (NMDeviceWifi *device,
 	if (!ssid || !nm_utils_same_ssid (nm_setting_wireless_get_ssid (s_wireless), ssid, TRUE))
 		return;
 
-	if (add_seen_bssid (gconf_connection, new))
-		nma_gconf_connection_save (gconf_connection);
+	add_seen_bssid (connection, new);
 
 	applet_schedule_update_icon (applet);
 }
 
-static guchar *
-ap_hash (NMAccessPoint * ap)
-{
-	struct GnomeKeyringMD5Context ctx;
-	unsigned char * digest = NULL;
-	unsigned char md5_data[66];
-	unsigned char input[33];
-	const GByteArray * ssid;
-	NM80211Mode mode;
-	guint32 flags, wpa_flags, rsn_flags;
-
-	g_return_val_if_fail (ap, NULL);
-
-	mode = nm_access_point_get_mode (ap);
-	flags = nm_access_point_get_flags (ap);
-	wpa_flags = nm_access_point_get_wpa_flags (ap);
-	rsn_flags = nm_access_point_get_rsn_flags (ap);
-
-	memset (&input[0], 0, sizeof (input));
-
-	ssid = nm_access_point_get_ssid (ap);
-	if (ssid)
-		memcpy (input, ssid->data, ssid->len);
-
-	if (mode == NM_802_11_MODE_INFRA)
-		input[32] |= (1 << 0);
-	else if (mode == NM_802_11_MODE_ADHOC)
-		input[32] |= (1 << 1);
-	else
-		input[32] |= (1 << 2);
-
-	/* Separate out no encryption, WEP-only, and WPA-capable */
-	if (  !(flags & NM_802_11_AP_FLAGS_PRIVACY)
-	    && (wpa_flags == NM_802_11_AP_SEC_NONE)
-	    && (rsn_flags == NM_802_11_AP_SEC_NONE))
-		input[32] |= (1 << 3);
-	else if (   (flags & NM_802_11_AP_FLAGS_PRIVACY)
-	         && (wpa_flags == NM_802_11_AP_SEC_NONE)
-	         && (rsn_flags == NM_802_11_AP_SEC_NONE))
-		input[32] |= (1 << 4);
-	else if (   !(flags & NM_802_11_AP_FLAGS_PRIVACY)
-	         &&  (wpa_flags != NM_802_11_AP_SEC_NONE)
-	         &&  (rsn_flags != NM_802_11_AP_SEC_NONE))
-		input[32] |= (1 << 5);
-	else
-		input[32] |= (1 << 6);
-
-	digest = g_malloc (sizeof (unsigned char) * AP_HASH_LEN);
-	if (digest == NULL)
-		goto out;
-
-	gnome_keyring_md5_init (&ctx);
-	memcpy (md5_data, input, sizeof (input));
-	memcpy (&md5_data[33], input, sizeof (input));
-	gnome_keyring_md5_update (&ctx, md5_data, sizeof (md5_data));
-	gnome_keyring_md5_final (digest, &ctx);
-
-out:
-	return digest;
-}
-
 static void
 add_hash_to_ap (NMAccessPoint *ap)
 {
-	guchar *hash = ap_hash (ap);
-	g_object_set_data_full (G_OBJECT (ap),
-	                        "hash", hash,
-	                        (GDestroyNotify) g_free);
+	char *hash;
+
+	hash = utils_hash_ap (nm_access_point_get_ssid (ap),
+	                      nm_access_point_get_mode (ap),
+	                      nm_access_point_get_flags (ap),
+	                      nm_access_point_get_wpa_flags (ap),
+	                      nm_access_point_get_rsn_flags (ap));
+	g_object_set_data_full (G_OBJECT (ap), "hash", hash, (GDestroyNotify) g_free);
 }
 
 static void
@@ -1052,7 +1063,9 @@ access_point_removed_cb (NMDeviceWifi *device,
 }
 
 static void
-on_new_connection (NMSettings *settings, NMExportedConnection *connection, gpointer datap)
+on_new_connection (NMSettingsInterface *settings,
+                   NMSettingsConnectionInterface *connection,
+                   gpointer datap)
 {
 	struct ap_notification_data *data = datap;
 	queue_avail_access_point_notification (NM_DEVICE (data->device));
@@ -1062,7 +1075,7 @@ static void
 free_ap_notification_data (gpointer user_data)
 {
 	struct ap_notification_data *data = user_data;
-	NMSettings *settings = applet_get_settings (data->applet);
+	NMSettingsInterface *settings = applet_get_settings (data->applet);
 
 	if (data->id)
 		g_source_remove (data->id);
@@ -1107,7 +1120,8 @@ wireless_device_added (NMDevice *device, NMApplet *applet)
 	 * that might be candididates.  Keep the ID around so we can disconnect
 	 * when the device is destroyed.
 	 */ 
-	id = g_signal_connect (applet_get_settings (applet), "new-connection",
+	id = g_signal_connect (applet_get_settings (applet),
+	                       NM_SETTINGS_INTERFACE_NEW_CONNECTION,
 	                       G_CALLBACK (on_new_connection),
 	                       data);
 	data->new_con_id = id;
@@ -1170,7 +1184,7 @@ wireless_device_state_changed (NMDevice *device,
                                NMDeviceStateReason reason,
                                NMApplet *applet)
 {
-	NMAGConfConnection *gconf_connection;
+	NMSettingsConnectionInterface *connection;
 	NMAccessPoint *new = NULL;
 	char *msg;
 	char *esc_ssid = NULL;
@@ -1190,9 +1204,9 @@ wireless_device_state_changed (NMDevice *device,
 			esc_ssid = nm_utils_ssid_to_utf8 ((const char *) ssid->data, ssid->len);
 
 		/* Save this BSSID to seen-bssids list */
-		gconf_connection = applet_get_exported_connection_for_device (device, applet);
-		if (gconf_connection && add_seen_bssid (gconf_connection, new))
-			nma_gconf_connection_save (gconf_connection);
+		connection = applet_get_exported_connection_for_device (device, applet);
+		if (connection)
+			add_seen_bssid (connection, new);
 	}
 
 	msg = g_strdup_printf (_("You are now connected to the wireless network '%s'."),
@@ -1256,15 +1270,15 @@ wireless_get_icon (NMDevice *device,
 			strength = CLAMP (strength, 0, 100);
 
 			if (strength > 80)
-				pixbuf = applet->wireless_100_icon;
+				pixbuf = nma_icon_check_and_load ("nm-signal-100", &applet->wireless_100_icon, applet);
 			else if (strength > 55)
-				pixbuf = applet->wireless_75_icon;
+				pixbuf = nma_icon_check_and_load ("nm-signal-75", &applet->wireless_75_icon, applet);
 			else if (strength > 30)
-				pixbuf = applet->wireless_50_icon;
+				pixbuf = nma_icon_check_and_load ("nm-signal-50", &applet->wireless_50_icon, applet);
 			else if (strength > 5)
-				pixbuf = applet->wireless_25_icon;
+				pixbuf = nma_icon_check_and_load ("nm-signal-25", &applet->wireless_25_icon, applet);
 			else
-				pixbuf = applet->wireless_00_icon;
+				pixbuf = nma_icon_check_and_load ("nm-signal-00", &applet->wireless_00_icon, applet);
 
 			*tip = g_strdup_printf (_("Wireless network connection '%s' active: %s (%d%%)"),
 			                        id, ssid, strength);
@@ -1311,6 +1325,17 @@ nag_dialog_response_cb (GtkDialog *nag_dialog,
 	}
 }
 
+static void
+update_cb (NMSettingsConnectionInterface *connection,
+           GError *error,
+           gpointer user_data)
+{
+	if (error) {
+		g_warning ("%s: failed to update connection: (%d) %s",
+		           __func__, error->code, error->message);
+	}
+}
+
 static void
 wireless_dialog_response_cb (GtkDialog *foo,
                              gint response,
@@ -1321,7 +1346,6 @@ wireless_dialog_response_cb (GtkDialog *foo,
 	NMConnection *connection = NULL, *fuzzy_match = NULL;
 	NMDevice *device = NULL;
 	NMAccessPoint *ap = NULL;
-	NMAGConfConnection *gconf_connection;
 	const char *service = NM_DBUS_SERVICE_USER_SETTINGS;
 
 	if (response != GTK_RESPONSE_OK)
@@ -1343,6 +1367,9 @@ wireless_dialog_response_cb (GtkDialog *foo,
 		}
 	}
 
+	/* nma_wireless_dialog_get_connection() returns a connection with the
+	 * refcount incremented, so the caller must remember to unref it.
+	 */
 	connection = nma_wireless_dialog_get_connection (dialog, &device, &ap);
 	g_assert (connection);
 	g_assert (device);
@@ -1353,10 +1380,11 @@ wireless_dialog_response_cb (GtkDialog *foo,
 		goto activate;
 	}
 
-	gconf_connection = nma_gconf_settings_get_by_connection (applet->gconf_settings, connection);
-	if (gconf_connection) {
+	if (NMA_IS_GCONF_CONNECTION (connection)) {
 		/* Not a new or system connection, save the updated settings to GConf */
-		nma_gconf_connection_save (gconf_connection);
+		nm_settings_connection_interface_update (NM_SETTINGS_CONNECTION_INTERFACE (connection),
+		                                         update_cb,
+		                                         NULL);
 	} else {
 		GSList *all, *iter;
 
@@ -1391,12 +1419,12 @@ wireless_dialog_response_cb (GtkDialog *foo,
 				}
 			}
 
-			/* Balance the caller of wireless_dialog_new () */
+			/* Balance nma_wireless_dialog_get_connection() */
 			g_object_unref (connection);
-
 			connection = g_object_ref (fuzzy_match);
 		} else {
 			/* Entirely new connection */
+			NMAGConfConnection *new_gconf_connection;
 			NMSettingConnection *s_con;
 			char *id;
 
@@ -1424,11 +1452,15 @@ wireless_dialog_response_cb (GtkDialog *foo,
 			}
 
 			/* Export it over D-Bus */
-			gconf_connection = nma_gconf_settings_add_connection (applet->gconf_settings, connection);
-			if (!gconf_connection) {
-				nm_warning ("Couldn't create other network connection.");
+			new_gconf_connection = nma_gconf_settings_add_connection (applet->gconf_settings, connection);
+			if (!new_gconf_connection) {
+				nm_warning ("Couldn't create new network connection.");
 				goto done;
 			}
+
+			/* Balance nma_wireless_dialog_get_connection() */
+			g_object_unref (connection);
+			connection = g_object_ref (new_gconf_connection);
 		}
 	}
 
@@ -1442,7 +1474,7 @@ activate:
 	                               applet);
 
 done:
-	/* Balance the caller of wireless_dialog_new () */
+	/* Balance nma_wireless_dialog_get_connection() */
 	if (connection)
 		g_object_unref (connection);
 
@@ -1476,7 +1508,6 @@ add_one_setting (GHashTable *settings,
                  GError **error)
 {
 	GHashTable *secrets;
-	GError *tmp_error = NULL;
 
 	g_return_val_if_fail (settings != NULL, FALSE);
 	g_return_val_if_fail (connection != NULL, FALSE);
@@ -1484,22 +1515,13 @@ add_one_setting (GHashTable *settings,
 	g_return_val_if_fail (error != NULL, FALSE);
 	g_return_val_if_fail (*error == NULL, FALSE);
 
-	if (!utils_fill_connection_certs (NM_CONNECTION (connection), &tmp_error)) {
-		g_set_error (error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INTERNAL_ERROR,
-					 "%s.%d (%s): failed to read connection certificates: (%d) %s.",
-					 __FILE__, __LINE__, __func__,
-					 tmp_error ? tmp_error->code : -1,
-					 tmp_error && tmp_error->message ? tmp_error->message : "(unknown)");
-		return FALSE;
-	}
-
 	secrets = nm_setting_to_hash (setting);
-	utils_clear_filled_connection_certs (connection);
-
 	if (secrets) {
 		g_hash_table_insert (settings, g_strdup (nm_setting_get_name (setting)), secrets);
 	} else {
-		g_set_error (error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INTERNAL_ERROR,
+		g_set_error (error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_INTERNAL_ERROR,
 		             "%s.%d (%s): failed to hash setting '%s'.",
 		             __FILE__, __LINE__, __func__, nm_setting_get_name (setting));
 	}
@@ -1512,7 +1534,8 @@ typedef struct {
 	NMActiveConnection *active_connection;
 	GtkWidget *dialog;
 	GtkWidget *nag_dialog;
-	DBusGMethodInvocation *context;
+	NMANewSecretsRequestedFunc callback;
+	gpointer callback_data;
 	char *setting_name;
 } NMWifiInfo;
 
@@ -1534,7 +1557,6 @@ get_secrets_dialog_response_cb (GtkDialog *foo,
 {
 	NMWifiInfo *info = user_data;
 	NMAWirelessDialog *dialog = NMA_WIRELESS_DIALOG (info->dialog);
-	NMAGConfConnection *gconf_connection;
 	NMConnection *connection = NULL;
 	NMSettingWirelessSecurity *s_wireless_sec;
 	NMDevice *device = NULL;
@@ -1567,7 +1589,9 @@ get_secrets_dialog_response_cb (GtkDialog *foo,
 	g_object_weak_unref (G_OBJECT (info->active_connection), destroy_wifi_dialog, info);
 
 	if (response != GTK_RESPONSE_OK) {
-		g_set_error (&error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_SECRETS_REQUEST_CANCELED,
+		g_set_error (&error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_SECRETS_REQUEST_CANCELED,
 		             "%s.%d (%s): canceled",
 		             __FILE__, __LINE__, __func__);
 		goto done;
@@ -1575,7 +1599,9 @@ get_secrets_dialog_response_cb (GtkDialog *foo,
 
 	connection = nma_wireless_dialog_get_connection (dialog, &device, NULL);
 	if (!connection) {
-		g_set_error (&error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INTERNAL_ERROR,
+		g_set_error (&error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_INTERNAL_ERROR,
 		             "%s.%d (%s): couldn't get connection from wireless dialog.",
 		             __FILE__, __LINE__, __func__);
 		goto done;
@@ -1584,7 +1610,9 @@ get_secrets_dialog_response_cb (GtkDialog *foo,
 	/* Second-guess which setting NM wants secrets for. */
 	s_wireless_sec = NM_SETTING_WIRELESS_SECURITY (nm_connection_get_setting (connection, NM_TYPE_SETTING_WIRELESS_SECURITY));
 	if (!s_wireless_sec) {
-		g_set_error (&error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INVALID_CONNECTION,
+		g_set_error (&error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_INVALID_CONNECTION,
 		             "%s.%d (%s): requested setting '802-11-wireless-security'"
 		             " didn't exist in the connection.",
 		             __FILE__, __LINE__, __func__);
@@ -1597,7 +1625,9 @@ get_secrets_dialog_response_cb (GtkDialog *foo,
 	settings = g_hash_table_new_full (g_str_hash, g_str_equal,
 	                                  g_free, (GDestroyNotify) g_hash_table_destroy);
 	if (!settings) {
-		g_set_error (&error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INTERNAL_ERROR,
+		g_set_error (&error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_INTERNAL_ERROR,
 		             "%s.%d (%s): not enough memory to return secrets.",
 		             __FILE__, __LINE__, __func__);
 		goto done;
@@ -1618,7 +1648,9 @@ get_secrets_dialog_response_cb (GtkDialog *foo,
 
 			s_8021x = (NMSetting8021x *) nm_connection_get_setting (connection, NM_TYPE_SETTING_802_1X);
 			if (!s_8021x) {
-				g_set_error (&error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INVALID_CONNECTION,
+				g_set_error (&error,
+				             NM_SETTINGS_INTERFACE_ERROR,
+				             NM_SETTINGS_INTERFACE_ERROR_INVALID_CONNECTION,
 				             "%s.%d (%s): requested setting '802-1x' didn't"
 				             " exist in the connection.",
 				             __FILE__, __LINE__, __func__);
@@ -1635,15 +1667,17 @@ get_secrets_dialog_response_cb (GtkDialog *foo,
 	if (!add_one_setting (settings, connection, NM_SETTING (s_wireless_sec), &error))
 		goto done;
 
-	dbus_g_method_return (info->context, settings);
+	info->callback (NM_SETTINGS_CONNECTION_INTERFACE (connection), settings, NULL, info->callback_data);
 
 	/* Save the connection back to GConf _after_ hashing it, because
 	 * saving to GConf might trigger the GConf change notifiers, resulting
 	 * in the connection being read back in from GConf which clears secrets.
 	 */
-	gconf_connection = nma_gconf_settings_get_by_connection (info->applet->gconf_settings, connection);
-	if (gconf_connection)
-		nma_gconf_connection_save (gconf_connection);
+	if (NMA_IS_GCONF_CONNECTION (connection)) {
+		nm_settings_connection_interface_update (NM_SETTINGS_CONNECTION_INTERFACE (connection),
+		                                         update_cb,
+		                                         NULL);
+	}
 
 done:
 	if (settings)
@@ -1651,7 +1685,7 @@ done:
 
 	if (error) {
 		g_warning ("%s", error->message);
-		dbus_g_method_return_error (info->context, error);
+		info->callback (NM_SETTINGS_CONNECTION_INTERFACE (connection), NULL, error, info->callback_data);
 		g_error_free (error);
 	}
 
@@ -1663,11 +1697,12 @@ done:
 
 static gboolean
 wireless_get_secrets (NMDevice *device,
-                      NMConnection *connection,
+                      NMSettingsConnectionInterface *connection,
                       NMActiveConnection *active_connection,
                       const char *setting_name,
                       const char **hints,
-                      DBusGMethodInvocation *context,
+                      NMANewSecretsRequestedFunc callback,
+                      gpointer callback_data,
                       NMApplet *applet,
                       GError **error)
 {
@@ -1676,7 +1711,9 @@ wireless_get_secrets (NMDevice *device,
 	const char *specific_object;
 
 	if (!setting_name || !active_connection) {
-		g_set_error (error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INTERNAL_ERROR,
+		g_set_error (error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_INTERNAL_ERROR,
 		             "%s.%d (%s): setting name and active connection object required",
 		             __FILE__, __LINE__, __func__);
 		return FALSE;
@@ -1684,7 +1721,9 @@ wireless_get_secrets (NMDevice *device,
 
 	specific_object = nm_active_connection_get_specific_object (active_connection);
 	if (!specific_object) {
-		g_set_error (error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INTERNAL_ERROR,
+		g_set_error (error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_INTERNAL_ERROR,
 		             "%s.%d (%s): could not determine AP for specific object",
 		             __FILE__, __LINE__, __func__);
 		return FALSE;
@@ -1693,9 +1732,11 @@ wireless_get_secrets (NMDevice *device,
 	info = g_malloc0 (sizeof (NMWifiInfo));
 
 	ap = nm_device_wifi_get_access_point_by_path (NM_DEVICE_WIFI (device), specific_object);
-	info->dialog = nma_wireless_dialog_new (applet, connection, device, ap);
+	info->dialog = nma_wireless_dialog_new (applet, NM_CONNECTION (connection), device, ap);
 	if (!info->dialog) {
-		g_set_error (error, NM_SETTINGS_ERROR, NM_SETTINGS_ERROR_INTERNAL_ERROR,
+		g_set_error (error,
+		             NM_SETTINGS_INTERFACE_ERROR,
+		             NM_SETTINGS_INTERFACE_ERROR_INTERNAL_ERROR,
 		             "%s.%d (%s): couldn't display secrets UI",
 		             __FILE__, __LINE__, __func__);
 		g_free (info);
@@ -1704,7 +1745,8 @@ wireless_get_secrets (NMDevice *device,
 
 	info->applet = applet;
 	info->active_connection = active_connection;
-	info->context = context;
+	info->callback = callback;
+	info->callback_data = callback_data;
 	info->setting_name = g_strdup (setting_name);
 
 	g_signal_connect (info->dialog, "response",
